@@ -53,7 +53,7 @@ except ImportError:
     import pickle
 
 __author__    = "Evan Martin <martine@danga.com>"
-__version__   = "1.2"
+__version__ = "1.31"
 __copyright__ = "Copyright (C) 2003 Danga Interactive"
 __license__   = "Python"
 
@@ -112,6 +112,37 @@ class Client:
         self.servers = [_Host(s, self.debuglog) for s in servers]
         self._init_buckets()
 
+    def get_stats(self):
+        '''Get statistics from each of the servers.  
+
+        @return: A list of tuples ( server_identifier, stats_dictionary ).
+            The dictionary contains a number of name/value pairs specifying
+            the name of the status field and the string value associated with
+            it.  The values are not converted from strings.
+        '''
+        data = []
+        for s in self.servers:
+            if not s.connect(): continue
+            name = '%s:%s (%s)' % ( s.ip, s.port, s.weight )
+            s.send_cmd('stats')
+            serverData = {}
+            data.append(( name, serverData ))
+            readline = s.readline
+            while 1:
+                line = readline()
+                if not line or line.strip() == 'END': break
+                stats = line.split(' ', 2)
+                serverData[stats[1]] = stats[2]
+
+        return(data)
+
+    def flush_all(self):
+        'Expire all data currently in the memcache servers.'
+        for s in self.servers:
+            if not s.connect(): continue
+            s.send_cmd('flush_all')
+            s.expect("OK")
+
     def debuglog(self, str):
         if self.debug:
             sys.stderr.write("MemCached: %s\n" % str)
@@ -137,8 +168,7 @@ class Client:
 
     def _get_server(self, key):
         if type(key) == types.TupleType:
-            serverhash = key[0]
-            key = key[1]
+            serverhash, key = key
         else:
             serverhash = hash(key)
 
@@ -367,8 +397,10 @@ class Client:
                 line = server.readline()
                 while line and line != 'END':
                     rkey, flags, rlen = self._expectvalue(server, line)
-                    val = self._recv_value(server, flags, rlen)
-                    retvals[rkey] = val
+                    #  Bo Yang reports that this can sometimes be None
+                    if rkey is not None:
+                        val = self._recv_value(server, flags, rlen)
+                        retvals[rkey] = val
                     line = server.readline()
             except (_Error, socket.error), msg:
                 server.mark_dead(msg)
@@ -402,20 +434,23 @@ class Client:
         elif flags & Client._FLAG_LONG:
             val = long(buf)
         elif flags & Client._FLAG_PICKLE:
-            val = pickle.loads(buf)
+            try:
+                val = pickle.loads(buf)
+            except:
+                self.debuglog('Pickle error...\n')
+                val = None
         else:
             self.debuglog("unknown flags on get: %x\n" % flags)
 
         return val
 
-    
+
 class _Host:
     _DEAD_RETRY = 30  # number of seconds before retrying a dead server.
 
     def __init__(self, host, debugfunc=None):
         if isinstance(host, types.TupleType):
-            host = host[0]
-            self.weight = host[1]
+            host, self.weight = host
         else:
             self.weight = 1
 
@@ -431,7 +466,7 @@ class _Host:
 
         self.deaduntil = 0
         self.socket = None
-    
+
     def _check_dead(self):
         if self.deaduntil and self.deaduntil > time.time():
             return 1
@@ -444,7 +479,7 @@ class _Host:
         return 0
 
     def mark_dead(self, reason):
-        print "MemCache: %s: %s.  Marking dead." % (self, reason)
+        self.debuglog("MemCache: %s: %s.  Marking dead." % (self, reason))
         self.deaduntil = time.time() + _Host._DEAD_RETRY
         self.close_socket()
         
@@ -469,26 +504,21 @@ class _Host:
             self.socket = None
 
     def send_cmd(self, cmd):
-        self.socket.sendall(cmd + "\r\n")
+        self.socket.sendall(cmd + '\r\n')
 
     def readline(self):
-        newlines = 0
-        buf = ''
-        while newlines < 2:
-            char = self.socket.recv(1) # XXX does this buffer or is this slow?
-            if len(char) == 0:
-                # connection closed.
-                print "MemCache: Connection closed while reading from %s.  Marking dead." % self
-                self.mark_dead
-                return buf
-            if char == '\r' and newlines == 0:
-                newlines = 1
-            elif char == '\n' and newlines == 1: 
-                newlines = 2
-            else:
-                newlines = 0
-                buf = buf + char
-        return buf
+        buffers = ''
+        recv = self.socket.recv
+        while 1:
+            data = recv(1)
+            if not data:
+                self.mark_dead('Connection closed while reading from %s'
+                        % repr(self))
+                break
+            if data == '\n' and buffers and buffers[-1] == '\r':
+                return(buffers[:-1])
+            buffers = buffers + data
+        return(buffers)
 
     def expect(self, text):
         line = self.readline()
@@ -498,8 +528,9 @@ class _Host:
     
     def recv(self, rlen):
         buf = ''
+        recv = self.socket.recv
         while len(buf) < rlen:
-            buf = buf + self.socket.recv(rlen - len(buf))
+            buf = buf + recv(rlen - len(buf))
         return buf
 
     def __str__(self):
